// alarm front panel
// -------------------------------------------------------------------
// based off of this code:
// LCD117_E
//
// copyright, Peter H Anderson, East Corinth, VT, July, '03
// set port to /dev/ttyUSB0
// set board to "Arduino Pro or Pro Mini"
// set processor to ATmega328 (5v, 16MHz)
// set programmer to USBTinyISP
//
// hacked by William F. Dudley Jr. for Arduino Pro-Mini 2015 01 25
//
// -------------------------------------------------------------------


//#define _19200
#define _9600
//#define _4800
//#define _2400
//#define _1200
//#define _300
#define SERIAL_DEBUG 1

#include <Arduino.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <HardwareSerial.h>
#include <EEPROM.h>
#include "EEPROM_rw_anything.h"
#include <Quadrature.h>

void lcd_init(void);
void lcd_clr(void);
void lcd_char(byte c);
void lcd_char_test(byte c);
void lcd_cmd_byte(byte c);
void lcd_data_nibble(byte c, byte delay);
void lcd_clr_line(byte line);
void lcd_cursor_pos(byte line, byte pos);
void lcd_line_feed(void);
void lcd_reverse_line_feed(void);
void lcd_CR(void);

void lcd_set_cursor_style(byte v);

void lcd_non_destructive_back_space(void);
void lcd_destructive_back_space(void);
void lcd_increment_cursor_pos(void);
void lcd_control_sequence(void);

void lcd_cursor_home(void);
void lcd_tab(void);
void lcd_beep(void);

void lcd_write_user_chars(void);
void lcd_write_big_num_patts(void);

void lcd_big3_num(char ch);
void lcd_big4_num(char ch);

void rlcd (char);

char num_to_char(byte val);

byte get_char(void);
byte ser_get_hex2(byte *p);

#define SUCCESS !0
#define FAILURE 0
#define TRUE !0
#define FALSE 0

#define CONFIG_REVISION_LEVEL 12326L
typedef struct config_t
{
  long magic_number;
  char custom_chars[8][8];
  uint8_t tab_size;
  uint8_t cursor;
  uint8_t num_lines;
  uint8_t num_cols;
  uint8_t backlight;
  uint8_t start_up;
  uint8_t custom_lines[4][20];
} CONFIG;

CONFIG config;

//#define DUMMY   // used for debugging

const  byte _0_mask[8] = {0xfe, 0xfd, 0xfb, 0xf7, 0xef, 0xdf, 0xbf, 0x7f};
const  byte _1_mask[8] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};

// for lcd
// #define BEEPER_PIN A3
#define BACKLIGHT_PIN 3
#define RS_PIN  2
#define ENABLE_PIN  4
#define FCK_LED 13

#define BUTTON1 A0
#define BUTTON2 A1
#define BUTTON3 12
#define BUTTON4  9
// this order matches the "1 2 3 4" membrane keypad
const byte buttons[] = { BUTTON1, BUTTON2, BUTTON3, BUTTON4 };
byte button_masks[4];
uint16_t button_mask_timers[4];
/* buttons are mapped to send 'A', 'B', 'C', and 'D'.
   also, buttons may be pressed two at a time (chording), so that:
   A+B = E, A+C = F, A+D = G
   B+C = H, B+D = I, C+D = J
   This mostly works correctly.
 */

// quadrature pot
#define QUAD_LIMIT_CONTROL 1
#define QD_PINA A3
#define QD_PINB A2
Quadrature quad1(QD_PINA, QD_PINB);

// #define AUX_PINS
// #define AUX4 A3
// #define AUX5 A4
// #define AUX6 A5

byte big_num3_flag;
byte big_num4_flag;

byte current_line, current_col;

void init_config (void) {
    if(config.magic_number != CONFIG_REVISION_LEVEL) { // set up first time
	config.magic_number = CONFIG_REVISION_LEVEL;
	memset((void *)&config.custom_chars[0][0],    0, 49);
	memset((void *)&config.custom_chars[6][1], 0x1f, 15);
	memset((void *)&config.custom_chars[5][2], 0x1f,  6);
	memset((void *)&config.custom_chars[4][3], 0x1f,  5);
	memset((void *)&config.custom_chars[3][4], 0x1f,  4);
	memset((void *)&config.custom_chars[2][5], 0x1f,  3);
	config.custom_chars[1][6] = 0x1f;
	config.custom_chars[1][7] = 0x1f;
	config.custom_chars[0][7] = 0x1f;
	config.tab_size  =    4;
	config.cursor    =    2;	// 3 is blinking block
	config.backlight = 0x80;
	config.num_lines =    4;
	config.num_cols  =   20;
	config.start_up  =    2;
#if OLD_CUSTOM
	for(uint8_t i = 0 ; i < 20 ; i++) {
	    config.custom_lines[0][i] = (uint8_t)('a' + i);
	    config.custom_lines[1][i] = (uint8_t)('A' + i);
	    config.custom_lines[2][i] = i % 10;
	    config.custom_lines[3][i] = (19-i) % 10;
	}
#else
	memcpy((void *)&config.custom_lines[0][0], " ALARM CONTROL PANEL", 20);
	for(uint8_t x = 1 ; x < 4 ; x++) {
	    memset((void *)&config.custom_lines[x][0], ' ', 20);
	}
#endif
	EEPROM_writeAnything(0, config);
    }

// custom_chars[0] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1f}
// custom_chars[1] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1f, 0x1f}
// custom_chars[2] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x1f, 0x1f, 0x1f}
// custom_chars[3] = {0x00, 0x00, 0x00, 0x00, 0x1f, 0x1f, 0x1f, 0x1f}
// custom_chars[4] = {0x00, 0x00, 0x00, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f}
// custom_chars[5] = {0x00, 0x00, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f}
// custom_chars[6] = {0x00, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f}
// custom_chars[7] = {0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f}

// custom_lines[0] = {"abcdefghijklmnopqrst"}
// custom_lines[1] = {"ABCDEFGHIJKLMNOPQRST"}
// custom_lines[2] = {"01234567890123456789"}
// custom_lines[3] = {"98765432109876543210"}
}

uint8_t eeprom_saved;
uint8_t old_q1, old_kb;

unsigned int no_key_found;

void setup(void) {
byte n, row, col;
    old_q1 = 0x80;
    old_kb = no_key_found = 0;
    Serial.begin(9600);
    init_config();
    eeprom_saved = 1;

    for(n = 0 ; n < sizeof(buttons) ; n++) {
	pinMode(buttons[n], INPUT_PULLUP);
	button_masks[n] = 0;
	button_mask_timers[n] = 0;
    }

#ifdef BEEPER_PIN
    pinMode(BEEPER_PIN, OUTPUT);
    digitalWrite(BEEPER_PIN, LOW);
#endif

#ifdef DEFER
    if (not_por == 0) {
	delay(300);
	not_por = 1;
	while(1)     ;  // wait for watch dog timeout
    }
#endif

    // test routine
    if (0) {
	delay(300); // ######## Nov 11
	while(1) {
#ifdef DEFER
	    if (not_por == 0) {
		not_por = 1;
		while(1)     ;  // wait for watch dog timeout
	    }
#endif
	    lcd_init();
	    lcd_cmd_byte(0x06);   // increment cursor
	    delay(2);
	    lcd_set_cursor_style(3);
	    for (n=32; n<128; n++) {
		if(n != '?') {
		    lcd_char_test(n);
		    delay(250);
		}
	    }
	}
    }
    big_num3_flag = FALSE;
    big_num4_flag = FALSE;

    delay(150);
    lcd_init();
    lcd_cmd_byte(0x06);   // increment cursor
    EEPROM_readAnything(0, config); // get saved settings
    lcd_set_cursor_style(config.cursor);
    lcd_write_user_chars();


    if (config.start_up ==1) {	// "stock" start screen
	char lcd_buff[8];
	lcd_clr_line(0);
#ifdef _19200
#ifdef SERIAL_DEBUG
	Serial.print("#117E 192 ");
#endif
	lcd_puts("#117E 192 ");
#endif
#ifdef _9600
#ifdef SERIAL_DEBUG
	Serial.print("#117E 96 ");
#endif
	lcd_puts("#117E 96 ");
#endif
#ifdef _4800
#ifdef SERIAL_DEBUG
	Serial.print("#117E 48 ");
#endif
	lcd_puts("#117E 48 ");
#endif
#ifdef _2400
#ifdef SERIAL_DEBUG
	Serial.print("#117E 24 ");
#endif
	lcd_puts("#117E 24 ");
#endif
	sprintf(lcd_buff, "%1dX%2d", config.num_lines, config.num_cols);
#ifdef SERIAL_DEBUG
	Serial.print(lcd_buff);
#endif
	lcd_puts(lcd_buff);
	lcd_clr_line(1);
	sprintf(lcd_buff, "%d ", config.tab_size);
#ifdef SERIAL_DEBUG
	Serial.print(lcd_buff);
#endif
	lcd_puts(lcd_buff);

	for (n=0; n<8; n++) { // display the special characters
	    lcd_char(n);
	}
	lcd_cursor_pos(1, 9);

	lcd_char(' ');

	analogWrite(BACKLIGHT_PIN, config.backlight);
	sprintf(lcd_buff, "%2x", config.backlight);
#ifdef SERIAL_DEBUG
	Serial.println(lcd_buff);
#endif
	lcd_puts(lcd_buff);
	delay(2000);
    }

    if (config.start_up  == 2) {	// custom start screen
	for (row = 0; row < config.num_lines; row++) {
	    lcd_clr_line(row);
	    for (col = 0; col < config.num_cols; col++) {
		 lcd_char(config.custom_lines[row][col]);
	    }
	}
	delay(2000);
    }
    lcd_clr();

    quad1.minimum(0);
    quad1.maximum(9);
}

#ifdef SERIAL_DEBUG
uint8_t linecount = 0;
#endif

void poll_serial(void) {
byte ch;
    if (Serial.available()) {
	ch = get_char();
	rlcd(ch);
    }
}

void my_delay(unsigned long td) {
unsigned long t0;
    t0 = td + millis();
    do {
	poll_serial();
    } while(millis() < t0);
}

void loop (void) {
byte ch, r, c;
uint8_t q1, qb, kb, n, m, kfound, found2;
    q1 = 0x80 | quad1.position();
    if(old_q1 != q1) {
        old_q1 = q1;
	qb = '0' + (q1 & 0x7f);
	// 20180102: this line was "missing", I can't figure out how pot got
	// sent to the alarm box without this code.
	Serial.write(qb);	// send pot position to master
#ifdef SERIAL_DEBUG
	linecount++;
	if(linecount > 39) {
	    linecount = 0;
	    Serial.write('\n');
	}
#endif
    }
    kfound = found2 = 0;
    for(n = 0 ; n < sizeof(buttons) ; n++) {
	if(button_masks[n]) {
	    button_mask_timers[n]--;
	    if(!button_mask_timers[n]) {
		button_masks[n] = 0;
	    }
	    else {
		continue;
	    }
	}
	ch = digitalRead(buttons[n]);
	if(!ch) {
	    // can't do chord w/last button as "first" button, so don't delay
	    if(n < (sizeof(buttons) - 1)) my_delay(100);
	    kfound++;
	    if(n < sizeof(buttons)) {
		for(m = n+1 ; m < sizeof(buttons) ; m++) {
		    if(button_masks[m]) { continue; }
		    if(!digitalRead(buttons[m])) {
			button_masks[n] = button_masks[m] = 1;
			button_mask_timers[n] = button_mask_timers[m] = 10000;
			kb = ((n == 0) ? 'D' : (n == 1) ? 'F' : 'G') + m;
			n = m;
			found2++;
		    }
		}
	    }
	    if(!found2) {
		kb = 'A' + n;
	    }
	    if(old_kb != kb) {
		old_kb = kb;
		Serial.write(kb);	// send key to master
#ifdef SERIAL_DEBUG
		linecount++;
		if(linecount > 39) {
		    linecount = 0;
		    Serial.write('\n');
		}
#endif
	    }
	}
    }
    if(!kfound) {
	no_key_found++;
	if(no_key_found > 1000) {
	    old_kb = 0;
	    no_key_found = 0;
	}
    }
    poll_serial();
    if(!eeprom_saved) {
	EEPROM_writeAnything(0, config);
	eeprom_saved = 1;
    }
}

void rlcd (char ch) {
    if (ch == 0xfe) {
	ch = get_char();
	lcd_cmd_byte(ch);
	delay(2);
    }
    else if (ch == '?') {
	lcd_control_sequence();
    }
    else if((ch >= 32) && (ch <= 128)) { // ignore special characters
	if (big_num3_flag == TRUE) {
	    if ((ch >= '0') && (ch <= '9')) {
		lcd_big3_num(ch); //<<<<<<<<<<<<<<<<<<<<<<<
	    }
	    else if ((ch == '.') || (ch==':') || (ch=='-')) {
		lcd_big3_num(ch); //<<<<<<<<<<<<<<<<<<<<<<<<<<<
	    }
	    else {
		// do nothing
	    }
	}
	else if (big_num4_flag == TRUE) {
	    if ((ch >= '0') && (ch <= '9')) {
		lcd_big4_num(ch);
	    }
	    else if ((ch == '.') || (ch==':') || (ch=='-')) {
		lcd_big4_num(ch);
	    }
	    else {
		// do nothing
	    }
	}
	else {
	    lcd_char(ch);
	    lcd_increment_cursor_pos();
	}
    }
}

void lcd_control_sequence (void) {
byte n, x, y, v, row, col;
byte ch, ch1, change;
byte vals[8];

    ch = get_char();

    if ((ch >= '0')  && (ch <='7')) {
	// write special character to LCD
	x = ch - '0';
	lcd_char(x);
	lcd_increment_cursor_pos();
    }
    else {
	switch(ch) {
	    case 'a':   // home cursor
		lcd_cursor_home();
		delay(2);
		break;

	    case 'b':   // destructive backspace
		lcd_destructive_back_space();
		break;

	    case 'c':   // cursor style
		ch1 = get_char();
		x = ch1 - '0';
		if ((x>=0) && (x<4)) {
		    lcd_set_cursor_style(x);
		    config.cursor = x;
		    eeprom_saved = 0;
		}
		break;
	    case 'f':
		lcd_clr();
		//delay(2);
		break;
	    case 'g':   // beep
		lcd_beep();
		break;


	    case 'G':   // set geometry
		ch1 = get_char();
		x = ch1 - '0';
		if ((x >= 2) && (x <=4)) { // config.num_lines
		    ch1 = get_char();   // first digit of config.num_cols
		    y = ch1 - '0';
		    if ((y>=1) && (y<=4)) {
			v = y;
		    }
		    else {
			break;
		    }

		    ch1 = get_char();   // second digit of config.num_cols
		    y = ch1 - '0';
		    if ((y>=0) && (y<=9)) {
			v = v * 10 + y;
		    }
		    else {
			break;
		    }
		    config.num_lines = x;
		    config.num_cols = v;
		    eeprom_saved = 0;
		}
		break;

	    case 'B':   // set backlight
		if (ser_get_hex2(&v) == SUCCESS) {
		    config.backlight = v;
		    analogWrite(BACKLIGHT_PIN, config.backlight);
		    eeprom_saved = 0;
		    break;
		}
		else {
		    break;
		}

	    case 'h':
		lcd_non_destructive_back_space();
		break;

	    case 'i':
		lcd_increment_cursor_pos();
		break;

	    case 'I':
		if (ser_get_hex2(&v) == SUCCESS) {
		    if ((v>=0) && (v<=3)) {
			lcd_char(v);
		    }
		}
		break;

	    case 'j':
		lcd_line_feed();
		break;

	    case 'k':
		lcd_reverse_line_feed();
		break;

	    case 'l':
		lcd_clr_line(current_line);
		break;

	    case 'm':
		lcd_CR();
		break;

	    case 'n':
		lcd_CR();
		lcd_line_feed();
		lcd_clr_line(current_line);
		break;

	    case 's':   // set horizonatal tab size
		ch1 = get_char();
		x = ch1 - '0';
		if ((x>0) && (x<9)) {
		    config.tab_size = x;
		    eeprom_saved = 0;
		}
		break;

	    case 't':
		lcd_tab();
		break;

	    case 'x':   // cursor_col
		ch1 = get_char();
		v = 0x00;
		x = ch1 - '0';
#ifdef OLD
		if ((x == 0) || (x == 1))
#endif
		if ((x>=0) && (x<=3)) {
		    v = x;
		    ch1 = get_char();
		    x = ch1 - '0';
		    if ((x>=0) && (x<=9)) {
			v = v*10 + x;
		    }
		    else {
			break;
		    }
		    lcd_cursor_pos(current_line, v);
		    break;
		}
		break;

	    case 'y':   // cursor_line
		ch1 = get_char();
		    x = ch1 - '0';
		    if ((x>=0) && (x<4)) {
		       lcd_cursor_pos(x, current_col);
		    }
		break;

	    case '!':   // command char to LCD
		if (ser_get_hex2(&v) == SUCCESS) {
		    lcd_cmd_byte(v);
		}
		break;


	    case '?':   lcd_char('?');
		lcd_increment_cursor_pos();
		break;

	    case 'D':   // define special character
		ch1 = get_char();
		x = ch1 - '0';
		if (!((x >=0) && (x<=7))) { // number is outside range
		    break;
		}

		for (n=0; n<8; n++) {
		    if (ser_get_hex2(&v) == SUCCESS) {
			vals[n] = v;
		    }
		    else {
			break;
		    }
		}

		// now save vals to EEPROM for special character x
		if(memcmp((void *)&config.custom_chars[x][0], vals, 8)) {
		    memcpy((void *)&config.custom_chars[x][0], vals, 8);
		    eeprom_saved = 0;
		}
		lcd_write_user_chars(); // also send to LCD RAM
		break;

#ifdef AUX_PINS
	    case 'H':	// high output on Aux pins 4,5, or 6
	    case 'L':	// low  output on Aux pins 4,5, or 6
		ch1 = get_char();
		x = ch1 - '0';
		if (!((x >=4) && (x<7))) { // number is outside range
		    break;
		}
		digitalWrite((x == 4) ? AUX4 : (x == 5) ? AUX5 : AUX6, (ch == 'H') ? HIGH : LOW);
#endif
#ifdef QUAD_LIMIT_CONTROL
	    case 'H':	// set quadrature pot max limit
	    case 'L':	// set quadrature pot min limit
		ch1 = get_char();
		if(ch1 >= '0' && ch1 <= '9') {
		    x = ch1 - '0';
		    if(ch == 'H') {
			quad1.maximum(x);
		    }
		    else {
			quad1.minimum(x);
		    }
		}
#endif
		break;
	    case 'Q':	// set current value of quad pot
		ch1 = get_char();
		if(ch1 >= '0' && ch1 <= '9') {
		    quad1.position(ch1 - '0');
		    old_q1 = 0x80 | (ch1 - '0');
		}
		break;
	    case 'S':
		ch1 = get_char();
		x = ch1 - '0';
		if (!((x >=0) && (x<=2))) { // number is outside range
		    break;
		}
		config.start_up = x;
		eeprom_saved = 0;
		break;
	    case '>': // turn on big characters
		ch1 = get_char();
		x = ch1 - '0';
		if (x == 3) {
		    big_num3_flag = TRUE;
		    big_num4_flag = FALSE;
		    lcd_write_big_num_patts();
		    break;
		}
		else if (x==4) {
		    big_num4_flag = TRUE;
		    big_num3_flag = FALSE;
		    lcd_write_big_num_patts();
		    break;
		}
		else {
		    break;
		}

	    case '<':
		big_num3_flag = FALSE;
		big_num4_flag = FALSE;
		break;

	    case 'R': // restore default custom characters
		lcd_write_user_chars();
		break;

	    case '*': // display custom screen, Mar, '05
		for (row = 0; row < config.num_lines; row++) {
		    lcd_clr_line(row);
		    for (col = 0; col < config.num_cols; col++) {
			lcd_char(config.custom_lines[row][col]);
		    }
		}
		current_line = 0; current_col = 0;
		break;

	    case 'C': // custom line definition, Mar, '05
		ch1 = get_char();
		change = 0;
		row = ch1 - '0';
		if (!((row >=0) && (row<=3))) { // number is outside range
		    break;
		}
		lcd_cursor_pos(row, 0);
		for (col=0 ; col<config.num_cols ; col++) {
		    ch = get_char();
		    if(config.custom_lines[row][col] != ch) {
			config.custom_lines[row][col] = ch;
			change = 1;
		    }
		    lcd_cursor_pos(row, col);
		    lcd_char(ch);
		}
		if(change) {
		    eeprom_saved = 0;
		}
		current_line = 0;  current_col = 0;
		lcd_cursor_pos(current_line, current_col);
		break;
	    default:
		break;
	}
    }
}

void lcd_big3_num(char ch) {

const byte _09[10][12] = {{4, 2, 5, 1, 0, 1, 1, 0, 1, 6, 3, 7},
			  {4, 1, 0, 0, 1, 0, 0, 1, 0, 3, 1, 3},
			  {4, 2, 5, 0, 4, 7, 4, 7, 0, 1, 3, 3},
			  {4, 2, 5, 0, 4, 7, 0, 6, 5, 6, 3, 7},
			  {0, 4, 1, 4, 7, 1, 2, 2, 1, 0, 0, 1},
			  {1, 2, 7, 1, 0, 0, 2, 2, 5, 6, 3, 7},
			  {4, 2, 5, 1, 0, 0, 1, 2, 5, 6, 3, 7},
			  {2, 2, 1, 0, 4, 7, 4, 7, 0, 1, 0, 0},
			  {4, 2, 5, 6, 3, 7, 4, 2, 5, 6, 3, 7},
			  {4, 2, 5, 1, 0, 1, 6, 2, 1, 6, 3, 7}};

byte col, row, num_index;

    if ((ch >='0') && (ch<='9')) {
	num_index = ch - '0';

	for (row=0; row<4; row++) {
	    lcd_cursor_pos(row, current_col);
	    for(col = 0; col < 4; col++) {
		if (col == 3) {
		    lcd_char(0);
		}
		else {
		    lcd_char(_09[num_index][3*row + col]);
		   //delay(250);
		   //lcd_increment_cursor_pos();
		}
	    }
	}
	current_col = current_col + 4;
    }
    else if (ch == '.') {
	for(row = 0; row<4; row++) {
	    lcd_cursor_pos(row, current_col);
	    for (col = 0; col < 2; col++) {
		if((col == 0) && (row == 3)) {
		    lcd_char(3);
		}
		else {
		    lcd_char(0);
		}
	    }
	}
	current_col = current_col + 2;
    }
    else if (ch == '-') {
	for(row = 0; row<4; row++) {
	    lcd_cursor_pos(row, current_col);
	    for (col = 0; col < 2; col++) {
		if((col == 0) && (row == 2)) {
		    lcd_char(2);
		}
		else {
		    lcd_char(0);
		}
	    }
	}
	current_col = current_col + 2;
    }

    else { // its a colon
	for(row = 0; row<4; row++) {
	    lcd_cursor_pos(row, current_col);
	    for (col = 0; col < 2; col++) {
		if((col == 0) && (row == 1)) {
		    lcd_char(2);
		}
		else if((col == 0) && (row == 2)) {
		    lcd_char(3);
		}
		else {
		    lcd_char(0);
		}
	    }
	}
	current_col = current_col + 2;
    }
    if (current_col > 19) {
	current_col = 0;
    }
    current_line = 0;
    lcd_cursor_pos(current_line, current_col);
}

void lcd_big4_num (char ch) {

const byte _09[10][16] = {{4, 2, 2, 5, 1, 0, 0, 1, 1, 0, 0, 1, 6, 3, 3, 7},
			  {0, 4, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 3, 1, 3},
			  {4, 2, 2, 5, 0, 0, 4, 7, 0, 4, 7, 0, 1, 3, 3, 3},
			  {4, 2, 2, 5, 0, 0, 4, 7, 0, 0, 6, 5, 6, 3, 3, 7},
			  {0, 4, 7, 1, 4, 7, 0, 1, 1, 3, 3, 1, 0, 0, 0, 1},
			  {1, 2, 2, 2, 6, 3, 3, 5, 0, 0, 0, 1, 6, 3, 3, 7},
			  {1, 2, 2, 0, 1, 3, 3, 5, 1, 0, 0, 1, 6, 3, 3, 7},
			  {2, 2, 2, 4, 0, 0, 4, 7, 0, 4, 7, 0, 1, 0, 0, 0},
			  {4, 2, 2, 5, 6, 3, 3, 7, 4, 2, 2, 5, 6, 3, 3, 7},
			  {4, 2, 2, 5, 1, 0, 0, 1, 6, 2, 2, 1, 0, 3, 3, 7}};

  // const b[16]
byte col, row, num_index;

    if ((ch >='0') && (ch<='9')) {
	num_index = ch - '0';

	for (row=0; row<4; row++) {
	    lcd_cursor_pos(row, current_col);
	    for(col = 0; col < 5; col++) {
		if (col == 4) {
		    lcd_char(0);
		}
		else {
		    lcd_char(_09[num_index][4*row + col]);
		    //delay(250);
		    //lcd_increment_cursor_pos();
		}
	    }
	}
	current_col = current_col + 5;
    }
    else if (ch == '.') {
	for(row = 0; row<4; row++) {
	    lcd_cursor_pos(row, current_col);
	    for (col = 0; col < 2; col++) {
		if((col == 0) && (row == 3)) {
		    lcd_char(3);
		}
		else {
		    lcd_char(0);
		}
	    }
	}
	current_col = current_col + 2;
    }
    else if (ch == '-') {
	for(row = 0; row<4; row++) {
	    lcd_cursor_pos(row, current_col);
	    for (col = 0; col < 2; col++) {
		if((col == 0) && (row == 2)) {
		    lcd_char(2);
		}
		else {
		    lcd_char(0);
		}
	    }
	}
	current_col = current_col + 2;
    }

    else { // its a colon
	for(row = 0; row<4; row++) {
	    lcd_cursor_pos(row, current_col);
	    for (col = 0; col < 2; col++) {
		if((col == 0) && (row == 1)) {
		    lcd_char(2);
		}
		else if((col == 0) && (row == 2)) {
		    lcd_char(3);
		}
		else {
		    lcd_char(0);
		}
	    }
	}
	current_col = current_col + 2;
    }
    if (current_col > 19) {
	current_col = 0;
    }
    current_line = 0;
    lcd_cursor_pos(current_line, current_col);
}

//-------------------foo-------------------
uint8_t _data_pins[4];

void lcd_my_init () {

    _data_pins[0] = 5;
    _data_pins[1] = 6;
    _data_pins[2] = 7;
    _data_pins[3] = 8;

    pinMode(BACKLIGHT_PIN, OUTPUT); // backlight
    analogWrite(BACKLIGHT_PIN, 0x80);
    for (uint8_t i=0; i<4; i++)
	pinMode(_data_pins[i], OUTPUT);
    pinMode(RS_PIN, OUTPUT);
    pinMode(ENABLE_PIN, OUTPUT);
}

void pulseEnable(void) {
    digitalWrite(ENABLE_PIN, LOW);
    delayMicroseconds(1);
    digitalWrite(ENABLE_PIN, HIGH);
    delayMicroseconds(1);    // enable pulse must be >450ns
    digitalWrite(ENABLE_PIN, LOW);
    delayMicroseconds(100);   // commands need > 37us to settle
}

void write4bits(uint8_t value) {
    for (int i = 0; i < 4; i++) {
	digitalWrite(_data_pins[i], (value >> i) & 0x01);
    }
    pulseEnable();
}


// LCD routines
void lcd_char(byte c) {    // displays ASCII character c to LCD
    lcd_data_nibble(c>>4, 0);    // high byte followed by low
    lcd_data_nibble(c&0x0f, 0);
    delayMicroseconds(70);
}

void lcd_puts(char *s) {    // displays ASCII string s to LCD
    for( ; *s ; s++) {
	lcd_char(*s);
    }
}

void lcd_char_test(byte c) {	// displays ASCII character c to LCD
    lcd_data_nibble(c>>4, 1);	// high byte followed by low
    lcd_data_nibble(c&0x0f, 1);
    delay(1);
}

void lcd_data_nibble(byte c, byte delay) { // RS is at logic one for data
    digitalWrite(RS_PIN, HIGH);
    write4bits(c);
}

void lcd_cmd_byte(byte c)  // used for sending byte commands
{
    digitalWrite(RS_PIN, LOW);
    write4bits(c>>4);    // high byte followed by low
    write4bits(c&0x0f);
    delayMicroseconds(70);
}

void lcd_init (void) {

    lcd_my_init();

// SEE PAGE 45/46 FOR INITIALIZATION SPECIFICATION!
// according to datasheet, we need at least 40ms after power rises above 2.7V
// before sending commands. Arduino can turn on way befer 4.5V so we'll wait 50
    delayMicroseconds(50000);
    // Now we pull both RS and R/W low to begin commands
    digitalWrite(RS_PIN, LOW);
    digitalWrite(ENABLE_PIN, LOW);

    // this is according to the hitachi HD44780 datasheet
    // figure 24, pg 46

    // we start in 8bit mode, try to set 4 bit mode
    write4bits(0x03);
    delayMicroseconds(4500); // wait min 4.1ms

    // second try
    write4bits(0x03);
    delayMicroseconds(4500); // wait min 4.1ms
  
    // third go!
    write4bits(0x03);
    delayMicroseconds(150);

    // finally, set to 8-bit interface
    write4bits(0x02);

    lcd_cmd_byte(0x01);
    delay(5);
    current_line = 0; current_col = 0;
}


void lcd_clr(void) {    // clear LCD and cursor to upper left
    lcd_cmd_byte(0x01);
    delay(2);
    current_line = 0; current_col = 0;
}

// clear indicated line and leave cursor at the beginning of the line
void lcd_clr_line(byte line) {
byte n;
    lcd_cursor_pos(line, 0);
    for (n=0; n<config.num_cols; n++) {
	lcd_char(' ');
    }
    lcd_cursor_pos(line, 0);
    current_line = line;  current_col = 0;
}

void lcd_cursor_pos(byte line, byte pos) {
const byte a[4] = {0x80, 0xc0, 0x80, 0xc0};
		    // start address for the two lines
    if ((line == 0) || (line == 1)) {
	lcd_cmd_byte(a[line]+pos);
    }
    else {
	lcd_cmd_byte(a[line] + config.num_cols + pos);
    }
    current_line = line;  current_col = pos;
}

void lcd_line_feed(void) {
   ++current_line;
   if (current_line > (config.num_lines - 1)) {
	current_line = 0;
   }
   lcd_cursor_pos(current_line, current_col);
}

void lcd_reverse_line_feed(void) {
    if (current_line == 0) {
	current_line = config.num_lines - 1;
    }
    else {
	--current_line;
    }
    lcd_cursor_pos(current_line, current_col);
}

void lcd_CR(void) {
   current_col = 0;
   lcd_cursor_pos(current_line, current_col);
}

void lcd_non_destructive_back_space(void) {
    if (current_col == 0) {
	if (current_line == 0) {
	    current_line = config.num_lines - 1;
	}
	else {
	    --current_line;
	}
	current_col = config.num_cols - 1;
    }
    else {
	-- current_col;
    }
    lcd_cursor_pos(current_line, current_col);
}

void lcd_destructive_back_space(void) {
    lcd_non_destructive_back_space();
    lcd_char(' ');
    lcd_increment_cursor_pos();
    lcd_non_destructive_back_space();
}

void lcd_set_cursor_style(byte v) {
    lcd_cmd_byte(0x0c+v);
}

void lcd_cursor_home(void) {
    lcd_cursor_pos(0, 0);
    current_line = 0;  current_col = 0;
}

void lcd_increment_cursor_pos(void) {
    ++current_col;
    if (current_col > (config.num_cols-1)) {
	current_col = 0;
	++current_line;
	if (current_line > (config.num_lines - 1)) {
	    current_line = 0;
	}
    }
    lcd_cursor_pos(current_line, current_col);
}

void lcd_tab(void) {
byte n;

    if (current_col >= (config.num_cols-1)) { // already at end of line
	// do nothing
    }
    else if (((current_col) % config.tab_size) == 0) { // currently at a tab set
	for (n=0; n<config.tab_size; n++) {
	    if (current_col >= (config.num_cols-1)) { // at end of line
		break;
	    }
	    lcd_char(' ');
	    lcd_increment_cursor_pos();
	}
    }
    else {
	for (n=0; ; n++) {
	    if (current_col >= (config.num_cols-1)) { // at end of line
		break;
	    }
	    else {
		if (((current_col) % config.tab_size) == 0) {
		    break;
		}
		lcd_char(' ');
		lcd_increment_cursor_pos();
	    }
	}
    }
}

void lcd_write_user_chars(void) { // write all user characters to LCD
byte char_num, n;

    for (char_num = 0; char_num < 8; char_num++) {
	// set CGRAM to address 0x00 + 8 * char_num
	lcd_cmd_byte(0x40 + 8 * char_num);
	for (n=0; n<8; n++) {
	    lcd_char(config.custom_chars[char_num][n]);
	}
	delay(5);
    }
    lcd_cursor_pos(current_line, current_col);   // return to regular mode
}

void lcd_write_big_num_patts(void) { // write big character patts to LCD
   byte char_num, n, x;
// big character patts
   const byte big[8][8] = {{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
			   {0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x00},
			   {0x1f, 0x1f, 0x1f, 0x1f, 0x00, 0x00, 0x00, 0x00},
			   {0x00, 0x00, 0x00, 0x1f, 0x1f, 0x1f, 0x1f, 0x00},
			   {0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x1f, 0x00},
			   {0x00, 0x10, 0x18, 0x1c, 0x1e, 0x1f, 0x1f, 0x00},
			   {0x1f, 0x1f, 0x0f, 0x07, 0x03, 0x01, 0x00, 0x00},
			   {0x1f, 0x1f, 0x1e, 0x1c, 0x18, 0x10, 0x00, 0x00}};


    for (char_num = 0; char_num < 8; char_num++) {
	lcd_cmd_byte(0x40 + 8 * char_num);   // set CGRAM to address 0x00 + 8 * char_num

	for (n=0; n<8; n++) {
	    x = big[char_num][n];
	    lcd_char(x);
	}
	delay(5);
    }
    lcd_cursor_pos(current_line, current_col);   // return to regular mode
}

void lcd_beep (void) {
#ifdef BEEPER_PIN
byte n;
    digitalWrite(BEEPER_PIN, LOW);

    for(n = 0; n<25; n++) {
	digitalWrite(BEEPER_PIN, HIGH);
	delayMicroseconds(1000);
	digitalWrite(BEEPER_PIN, LOW);
	delayMicroseconds(1000);
    }
#endif
}

///////////////// Serial Routines

byte get_char (void) {
    while(!Serial.available()) {
	if(!eeprom_saved) {
	    EEPROM_writeAnything(0, config);
	    eeprom_saved = 1;
	}
    }
    return Serial.read();
}

byte ser_get_hex2(byte *p) { // clean this up
byte ch, x, v;
    ch = get_char();
    if ((ch >= '0') && (ch<='9')) {
	x = ch - '0';
    }
    else if ((ch >='a') && (ch <='f')) {
	x = ch - 'a' + 10;
    }
    else if ((ch >='A') && (ch <='F')) {
	x = ch - 'A' + 10;
    }
    else {
	return(FAILURE);
    }
    v = x;
    ch = get_char();
    if ((ch >= '0') && (ch <='9')) {
	x = ch - '0';
    }
    else if ((ch >='a') && (ch <='f')) {
	x = ch - 'a' + 10;
    }
    else if ((ch >='A') && (ch <='F')) {
	x = ch - 'A' + 10;
    }
    else {
	return(FAILURE);
    }
    v = v*16 + x;
    *p = v;
    return(SUCCESS);
}

/* Baite BTE13-010 Pro Mini with CH340G usb to serial driver notes:
 * set port to /dev/ttyUSB0
 * set board to "Arduino Pro or Pro Mini"
 * set processor to ATmega328 (5v, 16MHz)
 * set programmer to USBTinyISP
 * Note: monitor mode works!
 *
 * Serial: 0 (RX) and 1 (TX). Used to receive (RX) and transmit (TX)
 * TTL serial data. These pins are connected to the TX-0 and RX-1 pins of
 * the six pin header.
 *
 * External Interrupts: 2 and 3. These pins can be configured to trigger
 * an interrupt on a low value, a rising or falling edge, or a change in
 * value. See the attachInterrupt() function for details.
 *
 * PWM: 3, 5, 6, 9, 10, and 11. Provide 8-bit PWM output with the
 * analogWrite() function.
 *
 * SPI: 10 (SS), 11 (MOSI), 12 (MISO), 13 (SCK). These pins support SPI
 * communication, which, although provided by the underlying hardware,
 * is not currently included in the Arduino language.
 *
 * LED: 13. There is a built-in LED connected to digital pin 13. When
 * the pin is HIGH value, the LED is on, when the pin is LOW, it's off.
 *
 * The Pro Mini has 8 analog inputs, each of which provide 10 bits of
 * resolution (i.e. 1024 different values). Four of them are on the headers
 * on the edge of the board; two (inputs 4 and 5) on holes in the interior
 * of the board. The analog inputs measure from ground to VCC. Additionally,
 * some pins have specialized functionality:
 *
 * I2C: A4 (SDA) and A5 (SCL). Support I2C (TWI) communication using
 * the Wire library.
 *
 * A6 and A7 are analog input only
 *
 * Reset. Bring this line LOW to reset the microcontroller. Typically
 * used to add a reset button to shields which block the one on the board.
 */
